/**
 *
 * @file
 *
 *  PLASMA is a software package provided by:
 *  University of Tennessee, US,
 *  University of Manchester, UK.
 *
 * @generated from compute/zgetri_aux.c, normal z -> c, Tue Oct 30 16:48:21 2018
 *
 **/

#include "plasma.h"
#include "plasma_async.h"
#include "plasma_context.h"
#include "plasma_descriptor.h"
#include "plasma_internal.h"
#include "plasma_types.h"
#include "plasma_workspace.h"

/***************************************************************************//**
 *
 * @ingroup plasma_getri
 *
 * Auxiliary routine for plasma_cgetri.
 *
 *******************************************************************************
 *
 * @param[in] n
 *          The number of columns of the matrix B. n >= 0.
 *
 * @param[in,out] pA
 *          On entry, the upper-triangular part contains the inverse of the
 *          U-factor, and the lower-triangular part contains the L-factor,
 *          both factors are computed by plasma_cgetrf.
 *          On exit, the inverse of L*U, overwriting the factors.
 *
 * @param[in] lda
 *          The leading dimension of the array A. lda >= max(1,k).
 *
 *******************************************************************************
 *
 * @retval PlasmaSuccess successful exit
 *
 *******************************************************************************
 *
 * @sa plasma_omp_cgetri
 * @sa plasma_cgetri
 * @sa plasma_dgetri
 * @sa plasma_sgetri
 *
 ******************************************************************************/
int plasma_cgetri_aux(int n, plasma_complex32_t *pA, int lda)
{
    // Get PLASMA context.
    plasma_context_t *plasma = plasma_context_self();
    if (plasma == NULL) {
        plasma_fatal_error("PLASMA not initialized");
        return PlasmaErrorNotInitialized;
    }

    // Check input arguments.
    if (n < 0) {
        plasma_error("illegal value of n");
        return -1;
    }
    if (lda < imax(1, n)) {
        plasma_error("illegal value of lda");
        return -3;
    }

    // quick return
    if (n == 0)
        return PlasmaSuccess;

    // Set tiling parameters.
    int nb = plasma->nb;

    // Create tile matrices.
    plasma_desc_t A;
    plasma_desc_t W;
    int retval;
    retval = plasma_desc_general_create(PlasmaComplexFloat, nb, nb,
                                        n, n, 0, 0, n, n, &A);
    if (retval != PlasmaSuccess) {
        plasma_error("plasma_desc_general_create() failed");
        return retval;
    }
    retval = plasma_desc_general_create(PlasmaComplexFloat, nb, nb,
                                        n, nb, 0, 0, n, nb, &W);
    if (retval != PlasmaSuccess) {
        plasma_error("plasma_desc_general_create() failed");
        plasma_desc_destroy(&A);
        return retval;
    }

    // Initialize sequence.
    plasma_sequence_t sequence;
    retval = plasma_sequence_init(&sequence);

    // Initialize request.
    plasma_request_t request;
    retval = plasma_request_init(&request);

    // asynchronous block
    #pragma omp parallel
    #pragma omp master
    {
        // Translate to tile layout.
        plasma_omp_cge2desc(pA, lda, A, &sequence, &request);

        // Call the tile async function.
        plasma_omp_cgetri_aux(A, W, &sequence, &request);

        // Translate back to LAPACK layout.
        plasma_omp_cdesc2ge(A, pA, lda, &sequence, &request);
    }
    // implicit synchronization

    // Free matrices in tile layout.
    plasma_desc_destroy(&A);
    plasma_desc_destroy(&W);

    // Return status.
    int status = sequence.status;
    return status;
}

/***************************************************************************//**
 *
 * @ingroup plasma_getri
 *
 *  Computes triangular solve.
 *  Non-blocking tile version of plasma_cgetri_aux().
 *  Operates on matrices stored by tiles.
 *  All matrices are passed through descriptors.
 *  All dimensions are taken from the descriptors.
 *  Allows for pipelining of operations at runtime.
 *
 *******************************************************************************
 *
 * @param[in] A
 *          Descriptor of the matrix.
 *
 * @param[in] W
 *          Workspace of dimension (n, nb).
 *
 * @param[in] sequence
 *          Identifies the sequence of function calls that this call belongs to
 *          (for completion checks and exception handling purposes).  Check
 *          the sequence->status for errors.
 *
 * @param[out] request
 *          Identifies this function call (for exception handling purposes).
 *
 * @retval void
 *          Errors are returned by setting sequence->status and
 *          request->status to error values.  The sequence->status and
 *          request->status should never be set to PlasmaSuccess (the
 *          initial values) since another async call may be setting a
 *          failure value at the same time.
 *
 *******************************************************************************
 *
 * @sa plasma_cgetri
 * @sa plasma_omp_cgetri
 * @sa plasma_omp_dgetri
 * @sa plasma_omp_sgetri
 *
 ******************************************************************************/
void plasma_omp_cgetri_aux(plasma_desc_t A, plasma_desc_t W,
                           plasma_sequence_t *sequence,
                           plasma_request_t *request)
{
    // Get PLASMA context.
    plasma_context_t *plasma = plasma_context_self();
    if (plasma == NULL) {
        plasma_error("PLASMA not initialized");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }

    // Check input arguments.
    if (plasma_desc_check(A) != PlasmaSuccess) {
        plasma_error("invalid A");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }
    if (plasma_desc_check(W) != PlasmaSuccess) {
        plasma_error("invalid W");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }
    if (sequence == NULL) {
        plasma_error("NULL sequence");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }
    if (request == NULL) {
        plasma_error("NULL request");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }

    // quick return
    if (A.n == 0)
      return;

    // Call the parallel function.
    plasma_pcgetri_aux(A, W, sequence, request);
}
